# -*- coding: utf-8 -*-

from AccessControl import Unauthorized

class PowerTokenSecurityError(Unauthorized):
    """General error for providing bad token"""

class PowerTokenConfigurationError(Exception):
    """General error for bad configuration of a Power Token object"""


# -*- coding: utf-8 -*-

from zExceptions import BadRequest

from zope.component import getUtility

from Products.Five import BrowserView
from Products.CMFCore.utils import getToolByName

from collective.powertoken.core.interfaces import IPowerTokenUtility

class ConsumePowerTokenView(BrowserView):
    
    def __call__(self, *args, **kw):
        form = self.request.form
        if not form.get('token') or not form.get('path'):
            raise BadRequest('You need to provide the token and the content path')
        
        path = form['path']
        token = form['token']
        
        target = self._getTarget(path)
        
        utility = getUtility(IPowerTokenUtility)
        return utility.consumeActions(target, token)
        
    def _getTarget(self, path):
        """
        Path must be always "folder/foo" or "/folder/foo", not including portal object id
        """
        portal = getToolByName(self.context, 'portal_url').getPortalObject()
        if path.startswith('/'):
            path = path[1:]
        return portal.unrestrictedTraverse(path)


class ConsumeFirstPowerTokenView(ConsumePowerTokenView):
    """Return only the first result from all executed token"""
    
    def __call__(self, *args, **kw):
        return ConsumePowerTokenView.__call__(self, *args, **kw)[0]



# -*- coding: utf-8 -*-

from persistent import Persistent
from zope.interface import implements
from collective.powertoken.core.interfaces import IPowerActionConfiguration

class TokenActionConfiguration(Persistent):
    implements(IPowerActionConfiguration)
    
    def __init__(self, type, roles=[], oneTime=True, unrestricted=False, username=None, **kwargs):
        self.type = type
        self.roles = roles
        self.oneTime = oneTime
        self.unrestricted = unrestricted
        self.username = username
        self.params = kwargs

    def __repr__(self):
        return "<TokenActionConfiguration type='%s', roles=%s (%s), username=%s, oneTime=%s, params=%s>" % (
                    self.type,
                    self.roles,
                    self.unrestricted and 'unrestricted' or 'restricted',
                    self.username,
                    self.oneTime,
                    self.params
                    )



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized

from zope.component import getUtility, ComponentLookupError
from zope.annotation.interfaces import IAnnotations

from collective.powertoken.core import config
from collective.powertoken.core.exceptions import PowerTokenSecurityError, PowerTokenConfigurationError
from collective.powertoken.core.interfaces import IPowerTokenUtility, IPowerTokenizedContent 
from collective.powertoken.core.tests.base import TestCase

class TestToken(TestCase):

    def afterSetUp(self):
        self.setRoles(('Manager', ))
        portal = self.portal
        portal.invokeFactory(type_name="Document", id="testdoc")
        doc = portal.testdoc
        doc.edit(title="A test document")
        self.utility = getUtility(IPowerTokenUtility)
        self.doc = portal.testdoc
        self.request = self.portal.REQUEST

    def test_enablePowerToken(self):
        token = self.utility.enablePowerToken(self.doc, 'foo')
        configuration = IAnnotations(self.doc)[config.MAIN_TOKEN_NAME]
        self.assertTrue(IPowerTokenizedContent.providedBy(self.doc))
        self.assertEquals(configuration.keys()[0], token)

    def test_addAction(self):
        self.assertRaises(PowerTokenConfigurationError, self.utility.addAction, self.doc, 'fooToken', 'fooAction')
        token = self.utility.enablePowerToken(self.doc, 'foo')
        self.utility.addAction(self.doc, token, 'fooAction')
        annotations = IAnnotations(self.doc)
        self.assertEqual(len(annotations[config.MAIN_TOKEN_NAME][token]), 2)
        self.assertEqual(annotations[config.MAIN_TOKEN_NAME][token][0].type, 'foo')
        self.assertEqual(annotations[config.MAIN_TOKEN_NAME][token][1].type, 'fooAction')

    def test_disablePowerTokens(self):
        token1 = self.utility.enablePowerToken(self.doc, 'foo')
        token2 = self.utility.enablePowerToken(self.doc, 'foo')
        self.assertTrue(IPowerTokenizedContent.providedBy(self.doc))
        self.utility.disablePowerTokens(self.doc)
        self.assertFalse(IPowerTokenizedContent.providedBy(self.doc))
        self.assertRaises(PowerTokenConfigurationError, self.utility.verifyToken, self.doc, token1, False)
        self.assertRaises(PowerTokenConfigurationError, self.utility.verifyToken, self.doc, token2, False)
        self.assertEqual(IAnnotations(self.doc).get(config.MAIN_TOKEN_NAME), None)

    def test_verifyToken(self):
        self.assertRaises(PowerTokenConfigurationError, self.utility.verifyToken, self.doc, 'foo', False)
        token = self.utility.enablePowerToken(self.doc, 'foo')
        self.assertFalse(self.utility.verifyToken(self.doc, 'foo', False))
        self.assertRaises(PowerTokenSecurityError, self.utility.verifyToken, self.doc, 'foo')
        self.assertTrue(self.utility.verifyToken(self.doc, token))
        annotations = IAnnotations(self.doc)
        del annotations[config.MAIN_TOKEN_NAME]
        self.assertRaises(KeyError, self.utility.verifyToken, self.doc, 'foo', False)

    def test_consumeToken(self):
        token1 = self.utility.enablePowerToken(self.doc, 'foo')
        self.utility.addAction(self.doc, token1, 'anotherFoo')
        token2 = self.utility.enablePowerToken(self.doc, 'foo')
        configuration = IAnnotations(self.doc)[config.MAIN_TOKEN_NAME]
        self.assertEqual(len(configuration), 2)
        self.assertEqual(len(self.utility.consumeToken(self.doc, token1)), 2)
        self.assertEqual(len(configuration), 1)
        self.utility.consumeToken(self.doc, token2)
        self.assertEqual(IAnnotations(self.doc).get(config.MAIN_TOKEN_NAME), None)
    
    def test_consumePermanentToken(self):
        """When oneTime is False for at leat one action"""
        token = self.utility.enablePowerToken(self.doc, 'foo', oneTime=False)
        configuration = IAnnotations(self.doc)[config.MAIN_TOKEN_NAME]
        self.assertEquals(len(configuration), 1)
        self.assertEquals(len(configuration[token]), 1)
        self.utility.consumeToken(self.doc, token)
        self.utility.consumeToken(self.doc, token)
        self.assertEquals(len(configuration), 1)
        # now add another non-permanent action
        self.utility.addAction(self.doc, token, 'anotherFoo')
        self.assertEquals(len(configuration), 1)
        self.assertEquals(len(configuration[token]), 2)
        self.utility.consumeToken(self.doc, token)
        self.assertEquals(len(configuration), 1)
        self.assertEquals(len(configuration[token]), 1)
        self.utility.consumeToken(self.doc, token)
        self.assertEquals(len(configuration), 1)
        self.assertEquals(len(configuration[token]), 1)

    def test_removeToken(self):
        token1 = self.utility.enablePowerToken(self.doc, 'foo')
        token2 = self.utility.enablePowerToken(self.doc, 'foo')
        configuration = IAnnotations(self.doc)[config.MAIN_TOKEN_NAME]
        self.assertEquals(len(configuration), 2)
        self.utility.removeToken(self.doc, token1)
        self.assertEqual(len(configuration), 1)
        self.utility.removeToken(self.doc, token2)
        self.assertEqual(IAnnotations(self.doc).get(config.MAIN_TOKEN_NAME), None)

    def test_consumeActions(self):
        token = self.utility.enablePowerToken(self.doc, 'foo')
        self.assertEqual(self.utility.consumeActions(self.doc, token),
                         [('http://nohost/plone/testdoc', 'foo', {})])
        self.assertRaises(PowerTokenConfigurationError, self.utility.consumeActions, self.doc, token)
        token = self.utility.enablePowerToken(self.doc, 'foo', aaa=5)
        self.assertEqual(self.utility.consumeActions(self.doc, token),
                         [('http://nohost/plone/testdoc', 'foo', {'aaa': 5})])
        token = self.utility.enablePowerToken(self.doc, 'fake', aaa=5)
        self.assertRaises(ComponentLookupError, self.utility.consumeActions, self.doc, token)

    def test_consumeActionsWithMultipleActions(self):
        token = self.utility.enablePowerToken(self.doc, 'foo', abc='text')
        self.utility.addAction(self.doc, token, 'foo')
        self.assertEqual(self.utility.consumeActions(self.doc, token),
                         [('http://nohost/plone/testdoc', 'foo', {'abc': 'text'}),
                          ('http://nohost/plone/testdoc', 'foo', {}),])
        self.assertRaises(PowerTokenConfigurationError, self.utility.consumeActions, self.doc, token)

    def test_unrestrictedUser(self):
        token = self.utility.enablePowerToken(self.doc, 'viewfoo') 
        has_role, username = self.utility.consumeActions(self.doc, token)[0]
        self.assertEqual(has_role, 0)
        self.logout()
        self.setRoles(('Anonymous', ))
        token = self.utility.enablePowerToken(self.doc, 'viewfoo', unrestricted=True)
        has_role, username = self.utility.consumeActions(self.doc, token)[0]
        self.assertEqual(has_role, 1)

    def test_usernameChange(self):
        token = self.utility.enablePowerToken(self.doc, 'viewfoo', roles='Member') 
        has_role, username = self.utility.consumeActions(self.doc, token)[0]
        self.assertEqual(username, 'test_user_1_')
        token = self.utility.enablePowerToken(self.doc, 'viewfoo', roles='Member', username='Unknow') 
        has_role, username = self.utility.consumeActions(self.doc, token)[0]
        self.assertEqual(username, 'Unknow')
        self.logout()
        self.setRoles(('Anonymous', ))
        token = self.utility.enablePowerToken(self.doc, 'viewfoo', roles='Member') 
        has_role, username = self.utility.consumeActions(self.doc, token)[0]
        self.assertEqual(username, '')
        token = self.utility.enablePowerToken(self.doc, 'viewfoo', roles='Member', username='Unknow') 
        has_role, username = self.utility.consumeActions(self.doc, token)[0]
        self.assertEqual(username, 'Unknow')


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestToken))
    return suite



# -*- coding: utf-8 -*-

from Products.Five import zcml
from Products.Five import fiveconfigure

#from Testing import ZopeTestCase as ztc

from Products.PloneTestCase import PloneTestCase as ptc
from Products.PloneTestCase.layer import onsetup

from zope.interface import implements
from zope.component import provideAdapter
from Products.CMFCore.interfaces import IContentish

from zope.publisher.interfaces.browser import IHTTPRequest
from collective.powertoken.core.interfaces import IPowerActionProvider

@onsetup
def setup_product():
    """Set up additional products and ZCML required to test this product.

    The @onsetup decorator causes the execution of this body to be deferred
    until the setup of the Plone site testing layer.
    """

    # Load the ZCML configuration for this package and its dependencies

    fiveconfigure.debug_mode = True
    import collective.powertoken.core
    zcml.load_config('configure.zcml', collective.powertoken.core)
    fiveconfigure.debug_mode = False

    # We need to tell the testing framework that these products
    # should be available. This can't happen until after we have loaded
    # the ZCML.

    #ztc.installPackage('collective.powertoken')
    
    provideAdapter(
            TestPowerActionProvider,
            (IContentish,
             IHTTPRequest),
            provides=IPowerActionProvider,
            name='foo'
        )
    provideAdapter(
            AdvancedTestPowerActionProvider,
            (IContentish,
             IHTTPRequest),
            provides=IPowerActionProvider,
            name='viewfoo'
        )

# The order here is important: We first call the deferred function and then
# let PloneTestCase install it during Plone site setup

setup_product()
#ptc.setupPloneSite(products=['collective.powertoken.core'])
ptc.setupPloneSite()


class TestPowerActionProvider(object):
    implements(IPowerActionProvider)
    
    def __init__(self, context, request):
        self.context = context
        self.request = request
    
    def doAction(self, action):
        return self.context.absolute_url(), action.type, action.params


class AdvancedTestPowerActionProvider(object):
    implements(IPowerActionProvider)
    
    def __init__(self, context, request):
        self.context = context
        self.request = request
    
    def doAction(self, action):
        member = self.context.portal_membership.getAuthenticatedMember()
        return (member.has_role('Abc', self.context), member.getId())


class TestCase(ptc.PloneTestCase):
    """Base class used for test cases
    """

class FunctionalTestCase(ptc.FunctionalTestCase):
    """Test case class used for functional (doc-)tests
    """



# -*- coding: utf-8 -*-

import uuid

from persistent.list import PersistentList
from persistent.dict import PersistentDict

from zope.interface import implements, alsoProvides, noLongerProvides
from zope.annotation.interfaces import IAnnotations
from zope.component import getMultiAdapter
from zope.component.interfaces import ComponentLookupError

from collective.powertoken.core.interfaces import IPowerTokenUtility, IPowerTokenizedContent
from collective.powertoken.core.interfaces import IPowerActionProvider
from collective.powertoken.core.tokenaction import TokenActionConfiguration
from collective.powertoken.core.exceptions import PowerTokenSecurityError, PowerTokenConfigurationError
from collective.powertoken.core import config

from AccessControl.User import SimpleUser, UnrestrictedUser
from AccessControl.SecurityManagement import newSecurityManager, setSecurityManager, getSecurityManager

from Products.CMFCore.utils import getToolByName

class UnrestrictedUser(UnrestrictedUser):
    """Unrestricted user that still has an id."""

    def getId(self):
        """Return the ID of the user."""
        return self.getUserName()


class PowerTokenUtility(object):
    """ Utility for manage power tokens on contents
    """

    implements(IPowerTokenUtility)
    
    def _generateNewToken(self):
        return str(uuid.uuid4())

    def enablePowerToken(self, content, type, roles=[], oneTime=True, unrestricted=False,
                         username=None, **kwargs):
        annotations = IAnnotations(content)
        
        if not annotations.get(config.MAIN_TOKEN_NAME):
            annotations[config.MAIN_TOKEN_NAME] = PersistentDict()
            alsoProvides(content, IPowerTokenizedContent)
            content.reindexObject(idxs=['object_provides'])
        powertokens = annotations[config.MAIN_TOKEN_NAME]

        token = self._generateNewToken()

        if powertokens.get(token) is not None:
            raise KeyError('Token %s already stored in object %s' % (token,
                                                                     '/'.join(content.getPhysicalPath())))
        
        powertokens[token] = PersistentList()
        self.addAction(content, token, type, roles=roles, oneTime=oneTime, unrestricted=unrestricted,
                       username=username, **kwargs)
        return token

    def addAction(self, content, token, type, roles=[], oneTime=True, unrestricted=False,
                  username=None, **kwargs):
        self.verifyToken(content, token, True)
        annotations = IAnnotations(content)
        powertokens = annotations[config.MAIN_TOKEN_NAME]
        actions = powertokens[token]
        action = TokenActionConfiguration(type, roles=roles, oneTime=oneTime,
                                          unrestricted=unrestricted, username=username, **kwargs)
        actions.append(action)
        return action

    def verifyToken(self, content, token, raiseOnError=True):
        if not IPowerTokenizedContent.providedBy(content):
            raise PowerTokenConfigurationError("Content %s isn't an IPowerTokenizedContent" % '/'.join(content.getPhysicalPath()))
        annotations = IAnnotations(content)
        try:
            powertokens = annotations[config.MAIN_TOKEN_NAME]
        except KeyError:
            raise KeyError("Content %s doesn't provide tokens" % '/'.join(content.getPhysicalPath()))

        if not token or powertokens.get(token) is None:
            if raiseOnError:
                raise PowerTokenSecurityError('Token verification failed')
            return False
        return True

    def disablePowerTokens(self, content):
        annotations = IAnnotations(content)
        del annotations[config.MAIN_TOKEN_NAME]
        noLongerProvides(content, IPowerTokenizedContent)
        content.reindexObject(idxs=['object_provides'])

    def consumeToken(self, content, token):
        self.verifyToken(content, token, True)
        annotations = IAnnotations(content)
        powertokens = annotations[config.MAIN_TOKEN_NAME]
        results = []
        for action in powertokens[token][:]:
            if action.oneTime:
                powertokens[token].remove(action)
            if len(powertokens[token])==0:
                del powertokens[token]
            if len(powertokens)==0:
                self.disablePowerTokens(content)
            results.append(action)
        return results

    def removeToken(self, content, token):
        self.verifyToken(content, token, True)
        annotations = IAnnotations(content)
        powertokens = annotations[config.MAIN_TOKEN_NAME]
        del powertokens[token]
        if len(powertokens)==0:
            self.disablePowerTokens(content)

    def consumeActions(self, content, token):
        self.verifyToken(content, token)
        actions = self.consumeToken(content, token)
        results = []

        for action in actions:
            action_type = action.type
            
            try:
                adapter = getMultiAdapter((content, content.REQUEST), IPowerActionProvider, name=action_type)
            except ComponentLookupError:
                raise ComponentLookupError('Cannot find a provider for performing action "%s" on %s' % (action_type,
                                                                                                        '/'.join(content.getPhysicalPath())))
            try:
                if action.roles or action.unrestricted:
                    acl_users = getToolByName(content, 'acl_users')
                    old_sm = getSecurityManager()
                    Cls = SimpleUser
                    if action.unrestricted:
                        Cls = UnrestrictedUser                        
                    tmp_user = Cls(action.username or old_sm.getUser().getId() or '', '', action.roles, '')
                    tmp_user = tmp_user.__of__(acl_users)
                    newSecurityManager(None, tmp_user)
                results.append(adapter.doAction(action))
            finally:
                if action.roles:
                    setSecurityManager(old_sm)
        return results



# -*- coding: utf-8 -*-

from zope.interface import Interface, Attribute

class IPowerTokenizedContent(Interface):
    """Marker interface for flagging content provided with a power token"""


class IPowerTokenUtility(Interface):
    """Interface for the Power Token utility"""
    
    def enablePowerToken(content, type, roles=[], oneTime=True, unrestricted=False, username=None, **kwargs):
        """
        Enable a Power Token on the content, registering an action type.
        
        Additional keyword arguments provided are stored in the action.
        
        Return the generated token.
        """

    def addAction(content, token, type, roles=[], oneTime=True, unrestricted=False, username=None, **kwargs):
        """
        Add another action to an existing Power Token enabled content
        
        Returns the new action IPowerActionConfiguration object
        """

    def disablePowerTokens(self, content):
        """Disable the Power Token feature on the object"""

    def verifyToken(content, token, raiseOnError=False):
        """
        Check if the content provide actions for the given token
        """
    
    def removeToken(content, token):
        """
        Remove a token from the object, not consuming it
        """
    
    def consumeToken(content, token):
        """
        Delete the given token action from the list
        
        Return the action deleted
        """
    
    def consumeActions(content, token):
        """
        Execute what to do on the content for a given token
        
        Returns the list of results of all actions in order (if any)
        """


class IPowerActionProvider(Interface):
    """Marker interface for an object able to perform action on a content"""
    
    def doAction(action):
        """Execute the action.
        
        Action is an object providing IPowerActionConfiguration interface
        """


class IPowerActionConfiguration(Interface):
    """An action configuration, taken from a Power Token"""

    type = Attribute("Kind on action")
    roles = Attribute("Run this action with those roles")
    oneTime = Attribute("True if the action can be executed one time, then is consumed. "
                        "False for action that not really consume the token")
    unrestricted = Attribute("For use UnrestrictedUser or not")
    username = Attribute("A custom userid to use for the temp user")
    params = Attribute("Additional parameters, used for performing the action")





from setuptools import setup, find_packages
import os, sys

version = '0.2.0'

tests_require = ['zope.testing', 'Products.PloneTestCase']

install_requires=['setuptools',]

if sys.version_info < (2, 6):
    install_requires.append('uuid')

setup(name='collective.powertoken.core',
      version=version,
      description="A mechanism for bypass Plone security, accessing specific "
                  "URLs while giving a secret token (core package)",
      long_description=open("README.txt").read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      # Get more strings from
      # http://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
        "Framework :: Plone",
        "Framework :: Plone :: 3.3",
        "Framework :: Plone :: 4.0",
        "Framework :: Plone :: 4.1",
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        ],
      keywords='plone security token plonegov',
      author='RedTurtle Technology',
      author_email='sviluppoplone@redturtle.it',
      url='http://plone.org/products/collective.powertoken.core',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective', 'collective.powertoken'],
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      tests_require=tests_require,
      extras_require=dict(tests=tests_require),
      entry_points="""
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )



# -*- coding: utf-8 -*-

from zope.component import getUtility, getMultiAdapter

from collective.powertoken.core.tests.base import TestCase
from collective.powertoken.core.interfaces import IPowerTokenUtility
from collective.powertoken.core.exceptions import PowerTokenConfigurationError

class TestTokenView(TestCase):

    def afterSetUp(self):
        self.setRoles(('Manager', ))
        portal = self.portal
        portal.invokeFactory(type_name="Document", id="testdoc")
        doc = portal.testdoc
        doc.edit(title="A test document")
        self.utility = getUtility(IPowerTokenUtility)
        self.doc = portal.testdoc
        self.request = self.portal.REQUEST
        self.logout()
        self.setRoles(('Anonymous', ))

    def test_givenParameters(self):
        from zExceptions import BadRequest
        view = getMultiAdapter((self.portal, self.request), name='consume-powertoken')
        self.assertRaises(BadRequest, view)
        self.request.form['token'] = 'foo'
        view = getMultiAdapter((self.portal, self.request), name='consume-powertoken')
        self.assertRaises(BadRequest, view)
        del self.request.form['token']
        self.request.form['path'] = '/foo'
        view = getMultiAdapter((self.portal, self.request), name='consume-powertoken')
        self.assertRaises(BadRequest, view)
        # ***
        self.request.form['token'] = 'foo'
        self.request.form['path'] = '/foo'
        view = getMultiAdapter((self.portal, self.request), name='consume-powertoken')
        self.assertRaises(KeyError, view)
        self.request.form['path'] = 'testdoc'
        view = getMultiAdapter((self.portal, self.request), name='consume-powertoken')
        self.assertRaises(PowerTokenConfigurationError, view)

    def test_powerActionCall(self):
        token = self.utility.enablePowerToken(self.doc, 'foo')
        self.request.form['token'] = token
        self.request.form['path'] = 'testdoc'
        view = getMultiAdapter((self.portal, self.request), name='consume-powertoken')
        self.assertEqual(view(),  [('http://nohost/plone/testdoc', 'foo', {})])

    def test_getTarget(self):
        view = getMultiAdapter((self.portal, self.request), name='consume-powertoken')
        self.assertRaises(KeyError, view._getTarget, 'aaa')
        self.assertEqual(view._getTarget('testdoc').absolute_url_path(), '/plone/testdoc')
        self.assertEqual(view._getTarget('/testdoc').absolute_url_path(), '/plone/testdoc')


class TestFirstTokenView(TestCase):

    def afterSetUp(self):
        self.setRoles(('Manager', ))
        portal = self.portal
        portal.invokeFactory(type_name="Document", id="testdoc")
        doc = portal.testdoc
        doc.edit(title="A test document")
        self.utility = getUtility(IPowerTokenUtility)
        self.doc = portal.testdoc
        self.request = self.portal.REQUEST
        self.logout()
        self.setRoles(('Anonymous', ))

    def test_firstPowerActionCall(self):
        token = self.utility.enablePowerToken(self.doc, 'foo')
        self.request.form['token'] = token
        self.request.form['path'] = 'testdoc'
        view = getMultiAdapter((self.portal, self.request), name='consume-powertoken-first')
        self.assertEqual(view(),  ('http://nohost/plone/testdoc', 'foo', {}))


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestTokenView))
    suite.addTest(makeSuite(TestFirstTokenView))
    return suite